/*
								+---------------------------------+
								|                                 |
								|        ***   Targa   ***        |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|             Tga.cpp             |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 06-08-2006
 *
 *	rewrote to C++, used templates and template functions to reduce code length
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include "Tga.h"

#pragma pack(1)

struct TTgaHeader {
	unsigned char n_id_length;
	unsigned char n_palette_type;
	unsigned char n_image_type;
	unsigned short n_first_color;
	unsigned short n_palette_colors;
	unsigned char n_palette_entry_size;
	unsigned short n_left;
	unsigned short n_top;
	unsigned short n_image_width;
	unsigned short n_image_height;
	unsigned char n_bpp;
	unsigned char n_descriptor_bits;
};

#pragma pack()

template <class T>
struct TColor {
	typedef T TDataType;
	inline unsigned __int32 operator()(const TDataType &r_t_color) const;
};

template <class TPaletteIndex>
struct TPaletteColor : public TColor<TPaletteIndex> {
protected:
	const unsigned __int32 *m_p_palette;
	int m_n_palette_entries;
	int m_n_offset;

public:
	TPaletteColor(const unsigned __int32 *p_palette, int n_palette_entries, int n_offset)
		:m_p_palette(p_palette), m_n_palette_entries(n_palette_entries), m_n_offset(n_offset)
	{
	}

	inline unsigned __int32 operator()(const TDataType &r_t_color) const
	{
		return (r_t_color + m_n_offset >= 0 &&
			r_t_color + m_n_offset < m_n_palette_entries)? m_p_palette[r_t_color + m_n_offset] : 0;
	}
};

struct T_RGB8_Color : public TColor<unsigned char[3]> {
	inline unsigned __int32 operator()(const TDataType &r_t_color) const
	{
		return ((unsigned __int32)r_t_color[0] << 16) | ((unsigned __int32)r_t_color[1] << 8) |
			(unsigned __int32)r_t_color[2];
	}
};

struct T_RGBA8_Color : public TColor<unsigned __int32> {
	inline unsigned __int32 operator()(const TDataType &r_t_color) const
	{
		return (r_t_color & 0xff00ff00) | ((r_t_color & 0xff0000) >> 16) |
			((r_t_color & 0xff) << 16);
	}
};

struct T_RGB5_Color : public TColor<unsigned short> {
	inline unsigned __int32 operator()(const TDataType &r_t_color) const
	{
		return n_5_to_8((r_t_color >> 10) & 0x1f) |
			  (n_5_to_8((r_t_color >> 5) & 0x1f) << 8) |
			  (n_5_to_8(r_t_color & 0x1f) << 16);
	}

protected:
	static inline int n_5_to_8(int n_x)
	{
		return (n_x * 255 / 31) & 0xff;
		// can be substituted with simple 3-bit shift, but that way 5-bit
		// white will be not converted to 8-bit white (rather to very light gray)
	}
};

/*
 *	template <class TColorStruct>
 *	static bool TGA_Read_RLE_Scanline(const TColorStruct &r_t_color_converter,
 *		unsigned __int32 *p_output, int n_width, FILE *p_fr)
 *		- function for reading run-length compressed scanlines
 *		- r_t_color_converter is object, used for color conversion
 */
template <class TColorStruct>
static bool TGA_Read_RLE_Scanline(const TColorStruct &r_t_color_converter,
	unsigned __int32 *p_output, int n_width, FILE *p_fr)
{
	for(int i = 0; i < n_width;) {
		unsigned char n_count;
		if(fread(&n_count, sizeof(char), 1, p_fr) != 1)
			return false;
		if(n_count & 0x80) {
			if(i + (n_count = (n_count & 0x7f) + 1) > n_width)
				return false;
			TColorStruct::TDataType t_value;
			if(fread(&t_value, sizeof(TColorStruct::TDataType), 1, p_fr) < 1)
				return false;
			for(unsigned __int32 *p_end = p_output + n_count, n_color =
			   r_t_color_converter(t_value); p_output < p_end;)
				*p_output ++ = n_color;
			// long block with constant color
		} else {
			if(i + ++ n_count > n_width)
				return false;
			TColorStruct::TDataType p_value[256], *p_cur_value = &p_value[0];
			if(fread(p_value, sizeof(TColorStruct::TDataType), n_count, p_fr) < n_count)
				return false;
			for(unsigned __int32 *p_end = p_output + n_count; p_output < p_end;)
				*p_output ++ = r_t_color_converter(*p_cur_value ++);
			// block with varying pixels
		}
		i += n_count;
	}

	return true;
}

/*
 *	template <class TColorStruct>
 *	static bool TGA_Read_Scanline(const TColorStruct &r_t_color_converter,
 *		unsigned __int32 *p_output, int n_width, FILE *p_fr)
 *		- function for reading uncompressed scanlines
 *		- r_t_color_converter is object, used for color conversion
 */
template <class TColorStruct>
static bool TGA_Read_Scanline(const TColorStruct &r_t_color_converter,
	unsigned __int32 *p_output, int n_width, FILE *p_fr)
{
	for(int i = 0, n_step = (n_width > 256)? 256 : n_width; i < n_width;
	   i += n_step, n_step = (n_width - i > 256)? 256 : n_width - i) {
		TColorStruct::TDataType p_value[256], *p_cur_value = &p_value[0];
		if(fread(p_value, sizeof(TColorStruct::TDataType), n_step, p_fr) < (unsigned)n_step)
			return false;
		for(unsigned __int32 *p_end = p_output + n_step; p_output < p_end;)
			*p_output ++ = r_t_color_converter(*p_cur_value ++);
		// block with varying pixels
	}

	return true;
}

enum {
	tga_Compressed_Mask = 8,
	tga_ImageType_Mask = 7,
	tga_ColorMapped = 1,
	tga_RGB = 2,
	tga_Grayscale = 3
};

TBmp *p_Load_TGA(const char *p_s_filename)
{
	FILE *p_fr;
	if(!(p_fr = fopen(p_s_filename, "rb")))
		return 0;
	// open file

	TTgaHeader t_header;
	if(fread(&t_header, sizeof(TTgaHeader), 1, p_fr) != 1 ||
	   fseek(p_fr, t_header.n_id_length, SEEK_CUR) != 0) {
		fclose(p_fr);
		return 0;
	}
	// read header

	if((t_header.n_image_type & tga_ImageType_Mask) == tga_Grayscale)
		t_header.n_palette_colors = 256; // f_ixme - is it already there? no, it isn't

	unsigned __int32 *p_palette;
	if(t_header.n_palette_colors > 0 &&
	   !(p_palette = new unsigned __int32[t_header.n_palette_colors])) {
		fclose(p_fr);
		return 0;
	}
	// alloc memory for palette

	if(t_header.n_palette_colors > 0 && (t_header.n_image_type & tga_ImageType_Mask) == tga_ColorMapped) {
		bool b_result;
		if(t_header.n_palette_entry_size == 16) {
			T_RGB5_Color t_color_converter;
			b_result = TGA_Read_Scanline<T_RGB5_Color>(t_color_converter, p_palette,
				t_header.n_palette_colors, p_fr);
		} else if(t_header.n_palette_entry_size == 24) {
			T_RGB8_Color t_color_converter;
			b_result = TGA_Read_Scanline<T_RGB8_Color>(t_color_converter, p_palette,
				t_header.n_palette_colors, p_fr);
		} else if(t_header.n_palette_entry_size == 32) {
			T_RGBA8_Color t_color_converter;
			b_result = TGA_Read_Scanline<T_RGBA8_Color>(t_color_converter, p_palette,
				t_header.n_palette_colors, p_fr);
		} else
			b_result = false;
		if(!b_result) {
			delete[] p_palette;
			fclose(p_fr);
			return 0;
		}
	}
	// read palette

	if((t_header.n_image_type & tga_ImageType_Mask) == tga_Grayscale) {
		for(unsigned __int32 *p_color = p_palette, *p_end = p_palette + 256, n_color = 0xff000000;
		   p_color < p_end; n_color += 0x00010101)
			*p_color ++ = n_color;
	}
	// grayscale pal

	TBmp *p_bitmap;
	if((t_header.n_image_type & tga_ImageType_Mask) < tga_ColorMapped ||
	   (t_header.n_image_type & tga_ImageType_Mask) > tga_Grayscale || !(p_bitmap = new TBmp)) {
		if(t_header.n_palette_colors > 0)
			delete[] p_palette;
		fclose(p_fr);
		return 0;
	}
	if(!(p_bitmap->p_buffer = new unsigned __int32[t_header.n_image_width *
	   t_header.n_image_height])) {
		if(t_header.n_palette_colors > 0)
			delete[] p_palette;
		delete p_bitmap;
		fclose(p_fr);
		return 0;
	}
	p_bitmap->n_width = t_header.n_image_width;
	p_bitmap->n_height = t_header.n_image_height;
	p_bitmap->b_grayscale = (t_header.n_image_type & tga_ImageType_Mask) == tga_Grayscale;
	p_bitmap->n_former_bpp = (p_bitmap->b_grayscale)? 8 : (((t_header.n_image_type & tga_ImageType_Mask) == tga_RGB)?
		t_header.n_bpp : t_header.n_palette_entry_size);
	p_bitmap->b_alpha = p_bitmap->n_former_bpp == 32;
	// alloc bitmap

	int n_step = (t_header.n_descriptor_bits & 0x20)? t_header.n_image_width :
		-t_header.n_image_width;
	for(unsigned __int32 *p_scanline = p_bitmap->p_buffer + ((t_header.n_descriptor_bits & 0x20)?
	   0 : t_header.n_image_width * (t_header.n_image_height - 1)), *p_end = p_bitmap->p_buffer +
	   ((t_header.n_descriptor_bits & 0x20)? t_header.n_image_width * t_header.n_image_height :
	   -t_header.n_image_width); p_scanline != p_end; p_scanline += n_step) {
		bool b_result;
		if((t_header.n_image_type & tga_ImageType_Mask) == tga_RGB) {
			if(t_header.n_image_type & tga_Compressed_Mask) {
				if(t_header.n_bpp == 16) {
					T_RGB5_Color t_color_converter;
					b_result = TGA_Read_RLE_Scanline<T_RGB5_Color>(t_color_converter, p_scanline,
						p_bitmap->n_width, p_fr);
				} else if(t_header.n_bpp == 24) {
					T_RGB8_Color t_color_converter;
					b_result = TGA_Read_RLE_Scanline<T_RGB8_Color>(t_color_converter, p_scanline,
						p_bitmap->n_width, p_fr);
				} else if(t_header.n_bpp == 32) {
					T_RGBA8_Color t_color_converter;
					b_result = TGA_Read_RLE_Scanline<T_RGBA8_Color>(t_color_converter, p_scanline,
						p_bitmap->n_width, p_fr);
				} else
					b_result = false;
			} else {
				if(t_header.n_bpp == 16) {
					T_RGB5_Color t_color_converter;
					b_result = TGA_Read_Scanline<T_RGB5_Color>(t_color_converter, p_scanline,
						p_bitmap->n_width, p_fr);
				} else if(t_header.n_bpp == 24) {
					T_RGB8_Color t_color_converter;
					b_result = TGA_Read_Scanline<T_RGB8_Color>(t_color_converter, p_scanline,
						p_bitmap->n_width, p_fr);
				} else if(t_header.n_bpp == 32) {
					T_RGBA8_Color t_color_converter;
					b_result = TGA_Read_Scanline<T_RGBA8_Color>(t_color_converter, p_scanline,
						p_bitmap->n_width, p_fr);
				} else
					b_result = false;
			}
		} else {
			if(t_header.n_image_type & tga_Compressed_Mask) {
				TPaletteColor<unsigned char> t_color_converter(p_palette,
					t_header.n_palette_colors, -t_header.n_first_color);
				b_result = TGA_Read_RLE_Scanline<TPaletteColor<unsigned char> >(t_color_converter,
					p_scanline, p_bitmap->n_width, p_fr);
			} else {
				TPaletteColor<unsigned char> t_color_converter(p_palette,
					t_header.n_palette_colors, -t_header.n_first_color);
				b_result = TGA_Read_Scanline<TPaletteColor<unsigned char> >(t_color_converter,
					p_scanline, p_bitmap->n_width, p_fr);
			}
		}
		if(!b_result) {
			delete[] p_bitmap->p_buffer;
			delete p_bitmap;
			p_bitmap = 0;
			break;
		}
	}

	if(t_header.n_palette_colors)
		delete[] p_palette;
	fclose(p_fr);
	// cleanup

	return p_bitmap;
}
